/*
 * $Id: $
 */
package com.neurothemis.xmlswarm.dom;

import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import goodsjpi.Persistent;
import goodslib.ArrayOfObject;


/**
 * PersistentNodeList implementation of w3c NodeList
 *
 * @author Guillaume BINET <gbin@netcourrier.com>
 * @version $Revision: $
 */
public class PersistentNodeList extends Persistent implements NodeList
{
   private static final boolean   doubleCapacity = true;
   private static final int      capacityGrowth = 10;

   private ArrayOfObject      nodes;
   private int                count;
   private transient int      cache;   // last accessed index

   /**
    * Default class constructor.
    */
   public PersistentNodeList ()
   {
      this(capacityGrowth);
   }

   /**
    * Class constructor with initial list capacity set
    * to given size.
    *
    * @param   initialCapacity
    *
    *      initial capacity of this list
    */
   public PersistentNodeList (int initialCapacity)
   {
      nodes = new ArrayOfObject(initialCapacity);
   }

   /**
    * Construct from another NodeList.
    */
   public PersistentNodeList (NodeList original, boolean clone)
   {
      int count = original.getLength();
      ArrayOfObject nodes = new ArrayOfObject(count);
      
      if (clone)
      {
         for (int i = 0; i < count; i++)
            nodes.putAt(i, (Persistent) original.item(i).cloneNode(true));
      }
      else
      {
         for (int i = 0; i < count; i++)
            nodes.putAt(i, (Persistent) original.item(i));
      }

      this.nodes = nodes;
      this.count = count;
   }

   public final Node getFirstNode ()
   {
      return count > 0 ? (Node) nodes.getAt(cache = 0) : null;
   }

   public final Node  getLastNode ()
   {
      return count > 0 ? (Node) nodes.getAt(cache = count - 1) : null;
   }

   public final int  getNodeIndex (Node node)
   {
      return findForward(node);
   }

   public final Node getNextNode (Node node)
   {
      int index = findForward(node);
      if (index >= 0 && ++index < count)
         return (Node) nodes.getAt(index);
      else
         return null;
   }

   public final Node getPreviousNode (Node node)
   {
      int index = findBackward(node);
      if (index > 0 && --index < count)
         return (Node) nodes.getAt(index);
      else
         return null;
   }

   public final void appendNode (Node node)
   {
      int newcount = count + 1;
      if (newcount > nodes.length())
         ensureCapacity(newcount);

      nodes.putAt(count++, (Persistent) node);
   }

   public final void insertNode (Node node, int index)
   {
      int newcount = count + 1;
      if (index < newcount)
      {
         if (newcount > nodes.length())
            ensureCapacity(newcount);

         nodes.copy (index+1,nodes.toArray(),index,count-index);
         nodes.putAt(index, (Persistent) node);
         count++;
      }
   }

   public final void replaceNode (Node node, int index)
   {
      if (index >= 0 && index < count)
         nodes.putAt(index, (Persistent) node);
   }

   public final void removeNode (Node node)
   {
      removeNode(getNodeIndex(node));
   }

   public final void removeNode (int index)
   {
      if (index < count && index >= 0)
      {
         int j = count - index - 1;
         if (j > 0)
            nodes.copy (index,nodes.toArray(),index+1,j);

         count--;
         nodes.putAt(count, (Persistent) null);
      }
   }

   public void clearNodes ()
   {
      
      int count = this.count;
      for (int i = 0; i < count; i++)
         nodes.putAt(i, null);
      this.count = 0;
      this.cache = 0;
   }

   public Node item(int index)
   {
      if (index >= 0 && index < count)
         return (Node) nodes.getAt(index);
      else
         return null;
   }

   public int getLength()
   {
      return count;
   }

   private final void ensureCapacity (int minCapacity)
   {
      int oldLength=nodes.length();
      
      int newLength;
      if (doubleCapacity)
         newLength = oldLength * 2;
      else
         newLength = oldLength + capacityGrowth;

      if (newLength < minCapacity)
         newLength = minCapacity;

      nodes.changeSize (newLength);
   }

   private final int findForward (Node node)
   {
      int   i;
      int count = this.count;
      int cache = this.cache;
      if (count > 0)
      {
         if (cache >= count)
            cache = count - 1;

        
         for (i = cache; i < count ; i++)
            if (node == nodes.getAt(i))
            {
               return this.cache = i;
            }

         if (cache > 0)
            for (i = 0; i < cache ; i++)
               if (node == nodes.getAt(i))
               {
                  return this.cache = i;
               }
      }
      this.cache = 0;
      return -1;
   }

   private final int findBackward (Node node)
   {
      int   i;
      int count = this.count;
      int cache = this.cache;
      if (count > 0)
      {
         if (cache >= count)
            cache = count - 1;

         for (i = cache; i >= 0; i--)
            if (node == nodes.getAt(i))
            {
               return this.cache = i;
            }

         if (++cache < count)
            for (i = cache; i < count; i++)
               if (node == nodes.getAt(i))
               {
                  return this.cache = i;
               }
      }
      this.cache = 0;
      return -1;
   }
}

