//-< CacheManager.java >---------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      4-Jan-2001  K.A. Knizhnik  * / [] \ *
//                          Last update:  4-Jan-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Object cache manager
//-------------------------------------------------------------------*--------*

package goodsjpi;

import java.util.Hashtable;

class CacheManager { 
    //
    // Critical section to synchronize access to static variables of metaobject
    //
    Object globalCS = new Object();
    // Header of l2-list of objects involved in trabsaction. Modified objects 
    // are placed at the beginning of the list (transObjects.next...), while
    // not-modified objects involved in transaction are placed at the end of
    // the list (transObjects.prev...)
    // This is done to optimize sending of transaction to the storage server
    // (no extra sort or repacking of objects is needed).
    //
    // If object was first included in transaction list as "non-modified" and
    // later was modified, then it will be relinked to the beginning of the 
    // transaction list. 
    //
    // Object are kept in the transaction list until the end of transaction to 
    // prevent GC from deallocating them before end of tranasction.
    //
    Persistent transObjects;
    //
    // Headers for two l2-lists used to implement enhanced LRU discipline
    // for replaing objects in object cache. Head of the list contains
    // most recently used object and tail - least recently used objects. 
    // To prevent object cache from replaing all object as a result of
    // database serach through large number of obejct (most of which will
    // not be used in future), standard LRU scheme was extended to more fit
    // specific of database applications. Object cache is divided into two 
    // part: primary cache and cache for frequently used objects. First time 
    // object is accessed, it is placed in primary cache. Object will be moved
    // to "frequently used objects" cache only when it is accessed more than
    // twice and it is not at the head of LRU list (case of several sucessive 
    // accesses to the object). Both cache parts are managed separately: when
    // total size of objects in the one part exceeds some limit value, least
    // recently used objects from this part are thrown away from clients memory
    // not affecting objects in another part of the cache.
    //
    Persistent lru0;
    Persistent lru1;
    //
    // Total sizes of objects in the part of the cache (see explanation above)
    // Stub objects are not placed in the cache so thier size in not included
    // into this values. The total size occpied by persistent objects in 
    // client memory therefore can be greater than cacheSize0+cacheSize1. 
    //
    int cacheSize0;
    int cacheSize1;
    //
    // Limits for total size of objects in both parts of the cache (see
    // explanation above). If size of some object exceed this limit values, 
    // it will be placed in the cache in any case. 
    //
    int cacheSizeLimit0 = 2*1024*1024;
    int cacheSizeLimit1 = 2*1024*1024;

    //
    // Minimal size of available free memory after reaching which object cache 
    // is cleaned to prevent hanging of thrown objects in the memory as
    // a result of presence references to this objects. 
    //
    long freeMemoryWatermark = 1024*1024;

    //
    // Counter of nested transactions (or nested invocations of methods of
    // classes derived from Persistent class). BasicMetaobject implements 
    // implicit scheme of detection transaction boundaries. Transaction is 
    // committed when no more active methods exists for the classes controlled 
    // by this metaobject (or metaobjects derived from Basicmetaobject), 
    // i.e. when nestedTransactionCount is equal to zero. It is possbile
    // to manualy increment and decrement this count, so explictly
    // set transaction boundaries (see methods beginNestedTransaction() and
    // endNestedTransaction() below.
    //
    int nestedTransactionCount;


    Persistent forgottenObjectList;

    public CacheManager() { 
	lru0 = new Persistent((Metaobject)null);
	lru0.next = lru0.prev = lru0;
	lru1 = new Persistent((Metaobject)null);
	lru1.next = lru1.prev = lru1;
	transObjects = new Persistent((Metaobject)null);
	transObjects.next = transObjects.prev = transObjects;
	forgottenObjectList = new Persistent((Metaobject)null);
	forgottenObjectList.next = forgottenObjectList.prev 
	    = forgottenObjectList;
    }

    static CacheManager defaultCacheManager = new CacheManager();
    static volatile int nAttached = 0; 
    static java.util.Hashtable cacheManagerHash = new java.util.Hashtable();

    public static CacheManager getCacheManager() { 
	if (nAttached == 0) { 
	    return defaultCacheManager;
	}
	CacheManager mng = (CacheManager)cacheManagerHash.get(Thread.currentThread());
	if (mng == null) { 
	    mng = defaultCacheManager;
	}
	return mng;
    }
    
    // 
    // Associate cache manager with thread
    //
    static final int hashCleanupPeriod = 64;

    public void attach() { 
	if (++nAttached % hashCleanupPeriod == 0) { 
	    java.util.Enumeration threads = cacheManagerHash.keys();
	    while (threads.hasMoreElements()) { 
		Thread thr = (Thread)threads.nextElement();
		if (!thr.isAlive()) { 
		    cacheManagerHash.remove(thr);
		}
	    }
	}
	cacheManagerHash.put(Thread.currentThread(), this);
    }

    //
    // Detach cache manager from thread 
    //
    public void detach() { 
	cacheManagerHash.remove(Thread.currentThread());
    }

    //
    // Manipulations with L2 list of objects
    //
    public static void linkAfter(Persistent after, Persistent obj) { 
	(obj.prev = after).next = (obj.next = after.next).prev = obj;
    }

    public static void linkBefore(Persistent before, Persistent obj) { 
	(obj.next = before).prev = (obj.prev = before.prev).next = obj;
    }

    public static void unlink(Persistent obj) { 
	Assert.that(obj.next.prev == obj && obj.prev.next == obj);
	obj.next.prev = obj.prev;
	obj.prev.next = obj.next;
	obj.next = obj.prev = null;
    }

    // Size of Persitent head plus size iused by finilizer and object hash entry  
    static final int objectAllocationOverhead = 64;

    //
    // Exclude object from LRU list so making it not available for 
    // cached objects LRU relpacing algorithm. This method is called
    // in two situations: 
    //  1) when object is accessed to prevent it from throwing away until
    //     it is accessed by active method.
    //  2) when modification notification for this object was received from
    //     server and instance of deteriorated object instance is removed from
    //     clients memory.
    //
    public synchronized void removeFromCache(Persistent obj) { 
	if ((obj.state & Persistent.CACHED) != 0) { 
	    Assert.that((obj.state
			 & (Persistent.DESTRUCTED|Persistent.RAW)) == 0);
	    unlink(obj);
	    if ((obj.state & Persistent.USEFUL) == 0) { 
		if ((obj.state & Persistent.ACCESSED) != 0) { 
		    if (obj != lru0.next) { 
			obj.state |= Persistent.USEFUL;
		    }
		} else { 
		    obj.state |= Persistent.ACCESSED;
		}
		cacheSize0 -= objectAllocationOverhead + obj.desc.objectSize(obj);
	    } else { 
		cacheSize1 -= objectAllocationOverhead + obj.desc.objectSize(obj);
	    }
	    obj.state &= ~Persistent.CACHED;
	}
    }

    //
    // Insert object in one of two cache LRU lists depending on its current
    // state. If total size of objects in this part of the cache exceeds limit 
    // value, least recently used objects will be removed from the cache
    // (and from client memory) until total size of objects becomes less or
    // equal than limit.
    //
    static Runtime runtime = Runtime.getRuntime();

    public synchronized void insertInCache(Persistent obj) { 
	Assert.that((obj.state 
		     & (Persistent.CACHED|Persistent.DESTRUCTED|Persistent.RAW)) == 0);
	long freeMemory = runtime.freeMemory();
	if ((obj.state & Persistent.USEFUL) == 0) { 
	    cacheSize0 += objectAllocationOverhead + obj.desc.objectSize(obj);
	    while ((freeMemory < freeMemoryWatermark
		    || cacheSize0 > cacheSizeLimit0) && lru0.prev != lru0) 
	    {
		Persistent victim = lru0.prev;
		cacheSize0 -= objectAllocationOverhead + victim.desc.objectSize(victim);
		victim.metaobject.destroyObject(victim);
		victim.state &= ~Persistent.CACHED;
		unlink(victim);
	    }
	    linkAfter(lru0, obj);
	} else { 
	    cacheSize1 += objectAllocationOverhead + obj.desc.objectSize(obj);
	    while ((freeMemory < freeMemoryWatermark 
		    || cacheSize1 > cacheSizeLimit1) && lru1.prev != lru1) 
	    {
		Persistent victim = lru1.prev;
		cacheSize1 -= objectAllocationOverhead + victim.desc.objectSize(victim);
		victim.metaobject.destroyObject(victim);
		victim.state &= ~Persistent.CACHED;
		unlink(victim);
	    }
	    linkAfter(lru1, obj);
	}
	obj.state |= Persistent.CACHED;
    }

    //
    // Include object in transaction list. If object is modified it will be 
    // placed at the head of the list, otherwise at the tail of the list.
    // If object already was included in transaction as "read-only", and 
    // later was modified, it will relinked to the head of the list.
    // 
    public void addToTransaction(Persistent obj) { 
	synchronized (transObjects) { 
	    if ((obj.state & Persistent.DIRTY) != 0) { 
		if ((obj.state & Persistent.TRANSWRITE) == 0) { 
		    if ((obj.state & Persistent.TRANSREAD) != 0) { 
			obj.state &= ~Persistent.TRANSREAD;
			unlink(obj);
		    }
		    linkAfter(transObjects, obj);
		    obj.state |= Persistent.TRANSWRITE;
		}
	    } else { 
		if ((obj.state & Persistent.TRANSREAD) == 0) { 
		    linkBefore(transObjects, obj);
		    obj.state |= Persistent.TRANSREAD;
		}
	    }
	}
    }

    //
    // Set new watermark for minimal size of fee memory
    //
    public void setFreeMemoryWatermark(long watermark) { 
	freeMemoryWatermark = watermark;
    }

    //
    // Set new limits for object cache. See description of cache management
    // algorithm at the beggining of this file.
    //
    public void setCacheLimits(int l0, int l1) { 
	synchronized (globalCS) { 
	    cacheSizeLimit0 = l0;
	    cacheSizeLimit1 = l1;
	}
    }

}
	
