package goodslib;

import goodsjpi.*;
import java.util.*;

/**
 * Hash table based implementation of the <tt>Map</tt> interface.  This
 * implementation provides all of the optional map operations, and permits
 * <tt>null</tt> values and the <tt>null</tt> key.  (The <tt>HashMap</tt>
 * class is roughly equivalent to <tt>Hashtable</tt>, except that it is
 * unsynchronized and permits nulls.)  This class makes no guarantees as to
 * the order of the map; in particular, it does not guarantee that the order
 * will remain constant over time.<p>
 *
 * This implementation provides constant-time performance for the basic
 * operations (<tt>get</tt> and <tt>put</tt>), assuming the hash function
 * disperses the elements properly among the buckets.  Iteration over
 * collection views requires time proportional to the "capacity" of the
 * <tt>HashMap</tt> instance (the number of buckets) plus its size (the number
 * of key-value mappings).  Thus, it's very important not to set the intial
 * capacity too high (or the load factor too low) if iteration performance is
 * important.<p>
 *
 * <b>Note that this implementation is not synchronized.</b> If multiple
 * threads access this map concurrently, and at least one of the threads
 * modifies the map structurally, it <i>must</i> be synchronized externally.
 * (A structural modification is any operation that adds or deletes one or
 * more mappings; merely changing the value associated with a key that an
 * instance already contains is not a structural modification.)  This is
 * typically accomplished by synchronizing on some object that naturally
 * encapsulates the map.  If no such object exists, the map should be
 * "wrapped" using the <tt>Collections.synchronizedMap</tt> method.  This is
 * best done at creation time, to prevent accidental unsynchronized access to
 * the map: <pre> Map m = Collections.synchronizedMap(new HashMap(...));
 * </pre><p>
 *
 * The iterators returned by all of this class's "collection view methods" are
 * <i>fail-fast</i>: if the map is structurally modified at any time after the
 * iterator is created, in any way except through the iterator's own
 * <tt>remove</tt> or <tt>add</tt> methods, the iterator will throw a
 * <tt>ConcurrentModificationException</tt>.  Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the
 * future.
 *
 * @author  Konstantin Knizhnik
 * @version 1.01, 10/06/00
 * @see     Object#hashCode()
 * @see     Collection
 * @see	    Map
 * @see	    TreeMap
 * @see	    Hashtable
 * @since 1.2
 */

public class PersistentHashMap extends PersistentAbstractMap 
implements Map, Cloneable 
{
    static class HtreePage extends Persistent {
	Persistent table[];
	int        used;

	static final int pageBits = 8;
	static final int pageSize = 1 << pageBits;
	
	void addChild(int i, Persistent child) { 
	    table[i & (pageSize-1)] = child;
	    used += 1;
	}

	void removeChild(int i, int bits, Entry e) { 
	    i &= (pageSize - 1);
	    if (table[i] == e) { 
		int mask = bits < 32 ? ((1 << bits) - 1) : -1;
		if (e.next != e && (e.next.hash & mask) == (e.hash & mask)) { 
		    table[i] = e.next;
		} else { 
		    table[i] = null;
		}
	    }
	    used -= 1;
	}

	HtreePage split(int i, int bits) { 
	    Persistent tab[] = new Persistent[pageSize];
	    i &= (pageSize - 1);
	    Entry first = (Entry)table[i], e = first;
	    int mask = (1 << bits) - 1;
	    int hash = e.hash & mask;
	    int n = 0;
	    do { 
		int h = (e.hash >>> bits) & (pageSize-1);
		Entry next = e.next;
		if (tab[h] != null) { 
		    e.unlink();
		    e.linkBefore((Entry)tab[h]);
		}
		tab[h] = e;
		e = next;
		n += 1;
	    } while (e != first && (e.hash & mask) == hash);
	    
	    used -= n - 1;
	    HtreePage pg = new HtreePage(tab, n);
	    table[i] = pg;
	    return pg;
	}

	void clear() { 
	    for (int i = table.length; --i >= 0; table[i] = null);
	    used = 0;
	}

	HtreePage() { 
	    table = new Persistent[pageSize];
	}

	HtreePage(Persistent tab[], int n) { 
	    table = tab;
	    used = n;
	}

	HtreePage(int i, Persistent entry) { 
	    used = 1;
	    table = new Persistent[pageSize];
	    table[i & (pageSize-1)] = entry;
	}
    }


    HtreePage root;
    int       count;

    /**
     * The table is rehashed when its size exceeds this threshold.  (The
     * value of this field is (int)(capacity * loadFactor).)
     *
     * @serial
     */
    int threshold;

    Entry entries;

    /**
     * The number of times this HashMap has been structurally modified
     * Structural modifications are those that change the number of mappings in
     * the HashMap or otherwise modify its internal structure (e.g.,
     * rehash).  This field is used to make iterators on Collection-views of
     * the HashMap fail-fast.  (See ConcurrentModificationException).
     */
    int modCount;

    /**
     * Constructs a new, empty map with the specified initial 
     * capacity and the specified load factor. 
     *
     * @param      initialCapacity   the initial capacity of the PersistentHashMap.
     * @param      loadFactor        the load factor of the PersistentHashMap
     * @throws     IllegalArgumentException  if the initial capacity is less
     *               than zero, or if the load factor is nonpositive.
     */
    public PersistentHashMap(int initialCapacity, float loadFactor) {
	if (initialCapacity < 0)
	    throw new IllegalArgumentException("Illegal Initial Capacity: "+
                                               initialCapacity);
        if (loadFactor <= 0 || Float.isNaN(loadFactor))
            throw new IllegalArgumentException("Illegal Load factor: "+
                                               loadFactor);
	root = new HtreePage();
	threshold = (int)(HtreePage.pageSize * loadFactor);
    }

    /**
     * Constructs a new, empty map with the specified initial capacity
     * and default load factor, which is <tt>0.75</tt>.
     *
     * @param   initialCapacity   the initial capacity of the PersistentHashMap.
     * @throws    IllegalArgumentException if the initial capacity is less
     *              than zero.
     */
    public PersistentHashMap(int initialCapacity) {
	this(initialCapacity, 1.75f);
    }

    /**
     * Constructs a new, empty map with a default capacity and load
     * factor, which is <tt>0.75</tt>.
     */
    public PersistentHashMap() {
	this(0, 1.75f);
    }

    /**
     * Constructs a new map with the same mappings as the given map.  The
     * map is created with a capacity of twice the number of mappings in
     * the given map or 11 (whichever is greater), and a default load factor,
     * which is <tt>0.75</tt>.
     *
     * @param t the map whose mappings are to be placed in this map.
     */
    public PersistentHashMap(Map t) {
	this(0, 1.75f);
	putAll(t);
    }

    /**
     * Returns the number of key-value mappings in this map.
     *
     * @return the number of key-value mappings in this map.
     */
    public int size() {
	return count;
    }

    /**
     * Returns <tt>true</tt> if this map contains no key-value mappings.
     *
     * @return <tt>true</tt> if this map contains no key-value mappings.
     */
    public boolean isEmpty() {
	return count == 0;
    }

    /**
     * Returns <tt>true</tt> if this map maps one or more keys to the
     * specified value.
     *
     * @param value value whose presence in this map is to be tested.
     * @return <tt>true</tt> if this map maps one or more keys to the
     *         specified value.
     */
    public boolean containsValue(Object value) {
	Entry first = entries;
	if (first == null) { 
	    return false;
	}
	Entry e = first;
	if (value == null) {
	    do { 
		if (e.value==null) { 
		    return true;
		} 
	    } while ((e = e.next) != first);
	} else {
	    do { 
		if (value.equals(e.value)) { 
		    return true;
		} 
	    } while ((e = e.next) != first);
	}
	return false;
    }

    /**
     * Returns <tt>true</tt> if this map contains a mapping for the specified
     * key.
     * 
     * @return <tt>true</tt> if this map contains a mapping for the specified
     * key.
     * @param key key whose presence in this Map is to be tested.
     */
    public boolean containsKey(Object key) {
	return getEntry(key) != null;
    }

    protected Entry getEntry(Object key) { 
	int hashCode = key.hashCode();
	int h = hashCode;
	HtreePage pg = root; 
	int bits = 0;
	while (true) { 
	    Persistent p = pg.table[h & (HtreePage.pageSize - 1)];
	    if (p == null) { 
		return null;
	    }
	    bits += HtreePage.pageBits;

	    if (p instanceof Entry) { 
		int mask = bits < 32 ? ((1 << bits) - 1) : -1;
		h = hashCode & mask;
		Entry e = (Entry)p; 
		do { 
		    if (e.hash == hashCode && e.key.equals(key)) { 
			return e;
		    }
		} while ((e = e.next) != p && (e.hash & mask) == h);	
		return null;
	    }
	    h >>>= HtreePage.pageBits;
	    pg = (HtreePage)p;
	}
    }

    /**
     * Returns the value to which this map maps the specified key.  Returns
     * <tt>null</tt> if the map contains no mapping for this key.  A return
     * value of <tt>null</tt> does not <i>necessarily</i> indicate that the
     * map contains no mapping for the key; it's also possible that the map
     * explicitly maps the key to <tt>null</tt>.  The <tt>containsKey</tt>
     * operation may be used to distinguish these two cases.
     *
     * @return the value to which this map maps the specified key.
     * @param key key whose associated value is to be returned.
     */
    public Object get(Object key) {
	Entry e = getEntry(key);
	return (e != null) ? e.value : null;
    }
	
    /**
     * Associates the specified value with the specified key in this map.
     * If the map previously contained a mapping for this key, the old
     * value is replaced.
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.  A <tt>null</tt> return can
     *	       also indicate that the PersistentHashMap previously associated
     *	       <tt>null</tt> with the specified key.
     */
    public Object put(Object key, Object value) {
	if (key instanceof String) { 
	    key = new ArrayOfChar((String)key);
	}
	Entry e;
	int hashCode = key.hashCode();
	int h = hashCode;
	HtreePage pg = root; 
	int bits = 0;
	while (true) { 
	    Persistent p = pg.table[h & (HtreePage.pageSize - 1)];
	    if (p == null) {
		entries = e = new Entry(key, value, hashCode, entries);
		pg.addChild(h, e);
		count += 1;
		modCount += 1;
		return null;
	    }
	    bits += HtreePage.pageBits;
	    if (p instanceof Entry) {
		int mask = bits < 32 ? ((1 << bits) - 1) : -1;
		e = (Entry)p;
		do { 
		    if (e.hash == hashCode && e.key.equals(key)) { 
			Object prevValue = e.value;
			e.value = (Persistent)value;
			return prevValue;
		    }		    
		} while ((e = e.next) != p && (e.hash & mask) == (hashCode & mask));

		if (pg.used > threshold && bits < 32) {  
		    // add new level to the branch
		    pg = pg.split(h, bits);
		    h >>>= HtreePage.pageBits;
		    p = pg.table[h & (HtreePage.pageSize - 1)];
		    if (p == null) { 
			p = entries;
		    }
		}
		entries = e = new Entry(key, value, hashCode, (Entry)p);
		count += 1;
		modCount += 1;
		pg.addChild(h, e);
	        return null;
	    }
	    h >>>= HtreePage.pageBits;
	    pg = (HtreePage)p;
	}
    }

    /**
     * Removes the mapping for this key from this map if present.
     *
     * @param key key whose mapping is to be removed from the map.
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.  A <tt>null</tt> return can
     *	       also indicate that the map previously associated <tt>null</tt>
     *	       with the specified key.
     */
    public Object remove(Object key) {
	int hashCode = key.hashCode();
	int h = hashCode;
	HtreePage pg = root; 
	int bits = 0;
	while (true) { 
	    Persistent p = pg.table[h & (HtreePage.pageSize - 1)];
	    if (p == null) { 
		return null;
	    }
	    bits += HtreePage.pageBits;

	    if (p instanceof Entry) { 
		int mask = bits < 32 ? ((1 << bits) - 1) : -1;
		Entry e = (Entry)p; 
		do {
		    if (e.hash == hashCode && e.key.equals(key)) { 
			modCount += 1;
			count -= 1;
			pg.removeChild(h, bits, e);
			if (e == entries) { 
			    entries = e.next == e ? null : e.next;
			}
			e.unlink();
			return e.value;
		    }
		} while ((e = e.next) != p && (e.hash & mask) == (hashCode & mask));	
		return null;
	    }
	    h >>>= HtreePage.pageBits;
	    pg = (HtreePage)p;
	}
    }

    /**
     * Copies all of the mappings from the specified map to this one.
     * 
     * These mappings replace any mappings that this map had for any of the
     * keys currently in the specified Map.
     *
     * @param t Mappings to be stored in this map.
     */
    public void putAll(Map t) {
	Iterator i = t.entrySet().iterator();
	while (i.hasNext()) {
	    Map.Entry e = (Map.Entry) i.next();
	    put(e.getKey(), e.getValue());
	}
    }

    /**
     * Removes all mappings from this map.
     */
    public void clear() {
	modCount++;
	root.clear();
	count = 0;
    }

    /**
     * Returns a shallow copy of this <tt>PersistentHashMap</tt> instance: the keys and
     * values themselves are not cloned.
     *
     * @return a shallow copy of this map.
     */
    public Object clone() {
	PersistentHashMap t;
	try { 
	    t = (PersistentHashMap)super.clone();
	} catch (CloneNotSupportedException e) { 
	    // this shouldn't happen, since we are Cloneable
	    throw new InternalError();
	}
	t.root = new HtreePage();
	t.entries  = null;
	t.keySet   = null;
	t.entrySet = null;
	t.values   = null;
	t.modCount = 0;
	t.putAll(this);
	return t;
    }


    public Entry getNext(HashIterator iterator) { 
	if (modCount != iterator.expectedModCount) { 
	    throw new ConcurrentModificationException();
	}
	Entry e = iterator.currEntry;
	if (e != null) { 
	    iterator.currEntry = e.next != iterator.firstEntry ? e.next : null;
	}
	return e;
    }

    public void removeCurr(HashIterator iterator) { 
	if (modCount != iterator.expectedModCount) { 
	    throw new ConcurrentModificationException();
	}
	remove(iterator.lastReturned.key);
	iterator.expectedModCount = modCount;
    }

    // Views

    private transient Set keySet = null;
    private transient Set entrySet = null;
    private transient Collection values = null;

    /**
     * Returns a set view of the keys contained in this map.  The set is
     * backed by the map, so changes to the map are reflected in the set, and
     * vice-versa.  The set supports element removal, which removes the
     * corresponding mapping from this map, via the <tt>Iterator.remove</tt>,
     * <tt>Set.remove</tt>, <tt>removeAll</tt>, <tt>retainAll</tt>, and
     * <tt>clear</tt> operations.  It does not support the <tt>add</tt> or
     * <tt>addAll</tt> operations.
     *
     * @return a set view of the keys contained in this map.
     */
    public Set keySet() {
	if (keySet == null) {
	    keySet = new AbstractSet() {
		public Iterator iterator() {
		    return getHashIterator(KEYS);
		}
		public int size() {
		    return count;
		}
                public boolean contains(Object o) {
                    return containsKey(o);
                }
		public boolean remove(Object o) {
                    int oldSize = count;
                    PersistentHashMap.this.remove(o);
		    return count != oldSize;
		}
		public void clear() {
		    PersistentHashMap.this.clear();
		}
	    };
	}
	return keySet;
    }

    /**
     * Returns a collection view of the values contained in this map.  The
     * collection is backed by the map, so changes to the map are reflected in
     * the collection, and vice-versa.  The collection supports element
     * removal, which removes the corresponding mapping from this map, via the
     * <tt>Iterator.remove</tt>, <tt>Collection.remove</tt>,
     * <tt>removeAll</tt>, <tt>retainAll</tt>, and <tt>clear</tt> operations.
     * It does not support the <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a collection view of the values contained in this map.
     */
    public Collection values() {
	if (values==null) {
	    values = new AbstractCollection() {
                public Iterator iterator() {
		    return getHashIterator(VALUES);
                }
                public int size() {
                    return count;
                }
                public boolean contains(Object o) {
                    return containsValue(o);
                }
                public void clear() {
                    PersistentHashMap.this.clear();
                }
            };
        }
	return values;
    }

    /**
     * Returns a collection view of the mappings contained in this map.  Each
     * element in the returned collection is a <tt>Map.Entry</tt>.  The
     * collection is backed by the map, so changes to the map are reflected in
     * the collection, and vice-versa.  The collection supports element
     * removal, which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Collection.remove</tt>,
     * <tt>removeAll</tt>, <tt>retainAll</tt>, and <tt>clear</tt> operations.
     * It does not support the <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a collection view of the mappings contained in this map.
     * @see Map.Entry
     */
    public Set entrySet() {
	if (entrySet==null) {
	    entrySet = new AbstractSet() {
                public Iterator iterator() {
		    return getHashIterator(ENTRIES);
                }

                public boolean contains(Object o) {
                    if (!(o instanceof Map.Entry)) { 
                        return false;
		    }
		    Entry e = getEntry(((Map.Entry)o).getKey());
		    return (e != null && e.equals(o));
		}

		public boolean remove(Object o) {		    
                    if (!(o instanceof Map.Entry)) { 
                        return false;
		    }
		    Object key = ((Map.Entry)o).getKey();
		    Entry e = getEntry(key);
		    if (e != null && e.equals(o)) { 
			remove(key);
			return true;
		    }
		    return false;
		}

                public int size() {
                    return count;
                }

                public void clear() {
                    PersistentHashMap.this.clear();
                }
            };
        }

	return entrySet;
    }

    private Iterator getHashIterator(int type) {
	if (count == 0) {
	    return emptyHashIterator;
	} else {
	    return new HashIterator(type);
	}
    }

    /**
     * PersistentHashMap collision list entry.
     */
    static class Entry extends Persistent implements Map.Entry {
	int        hash;
	Persistent key;
	Persistent value;
	Entry      next;
	Entry      prev;

	Entry(Object key, Object value, int hash, Entry node) {
	    this.hash  = hash;
	    this.key   = (Persistent)key;
	    this.value = (Persistent)value;
	    if (node != null) { 
		linkBefore(node);
	    } else { 
		next = prev = this;
	    }
	}

	void linkBefore(Entry node) { 
	    next = node;
	    prev = node.prev;
	    prev.next = node.prev = this;
	}

	void unlink() { 
	    next.prev = prev;
	    prev.next = next;
	    next = prev = null;
	}

	// Map.Entry Ops 

	public Object getKey() {
	    return key;
	}

	public Object getValue() {
	    return value;
	}

	public Object setValue(Object value) {
	    Object oldValue = this.value;
	    this.value = (Persistent)value;
	    return oldValue;
	}

	public boolean equals(Object o) {
	    if (!(o instanceof Map.Entry))
		return false;
	    Map.Entry e = (Map.Entry)o;

	    return (key==null ? e.getKey()==null : key.equals(e.getKey())) &&
	       (value==null ? e.getValue()==null : value.equals(e.getValue()));
	}

	public int hashCode() {
	    return hash ^ (value==null ? 0 : value.hashCode());
	}

	public String toString() {
	    return key+"="+value;
	}

	static Metaobject defaultMetaobject = new OptimisticMetaobject();
    }

    // Types of Iterators
    private static final int KEYS = 0;
    private static final int VALUES = 1;
    private static final int ENTRIES = 2;

    private static EmptyHashIterator emptyHashIterator 
	= new EmptyHashIterator();
					     
    private static class EmptyHashIterator implements Iterator {
	
	EmptyHashIterator() {
	    
	}

	public boolean hasNext() {
	    return false;
	}

	public Object next() {
	    throw new NoSuchElementException();
	}
	
	public void remove() {
	    throw new IllegalStateException();
	}

    }			
		    
    private class HashIterator implements Iterator {
	Entry currEntry;
	Entry firstEntry;
	Entry lastReturned;
	int type;

	/**
	 * The modCount value that the iterator believes that the backing
	 * List should have.  If this expectation is violated, the iterator
	 * has detected concurrent modification.
	 */
	private int expectedModCount = modCount;

	HashIterator(int type) {
	    currEntry = firstEntry = entries;
	    this.type = type;
	}

	public boolean hasNext() {
	    return currEntry != null;
	}

	public Object next() {
	    Entry e = getNext(this);
	    if (e != null) {
		lastReturned = e;
		return type == KEYS ? e.key : (type == VALUES ? e.value : e);
	    }
	    throw new NoSuchElementException();
	}

	public void remove() {
	    if (lastReturned == null) { 
		throw new IllegalStateException();
	    } 
	    if (lastReturned == firstEntry) { 
		firstEntry = firstEntry.next;
	    }
	    removeCurr(this);
	    lastReturned = null;
	}
    }
}



