package goodslib;

import goodsjpi.*;
import java.util.*;

/**
 * B-Tree tree based implementation of the <tt>SortedMap</tt> interface.
 * This class guarantees that the map will be in ascending key order, sorted
 * according to the <i>natural order</i> for the key's class (see
 * <tt>Comparable</tt>), or by the comparator provided at creation time,
 * depending on which constructor is used.<p>
 *
 * This implementation provides guaranteed log(n) time cost for the
 * <tt>containsKey</tt>, <tt>get</tt>, <tt>put</tt> and <tt>remove</tt>
 * operations.  Algorithms are adaptations of those in Cormen, Leiserson, and
 * Rivest's <I>Introduction to Algorithms</I>.<p>
 *
 * Note that the ordering maintained by a sorted map (whether or not an
 * explicit comparator is provided) must be <i>consistent with equals</i> if
 * this sorted map is to correctly implement the <tt>Map</tt> interface.  (See
 * <tt>Comparable</tt> or <tt>Comparator</tt> for a precise definition of
 * <i>consistent with equals</i>.)  This is so because the <tt>Map</tt>
 * interface is defined in terms of the equals operation, but a map performs
 * all key comparisons using its <tt>compareTo</tt> (or <tt>compare</tt>)
 * method, so two keys that are deemed equal by this method are, from the
 * standpoint of the sorted map, equal.  The behavior of a sorted map
 * <i>is</i> well-defined even if its ordering is inconsistent with equals; it
 * just fails to obey the general contract of the <tt>Map</tt> interface.<p>
 *
 * <b>Note that this implementation is not synchronized.</b> If multiple
 * threads access a map concurrently, and at least one of the threads modifies
 * the map structurally, it <i>must</i> be synchronized externally.  (A
 * structural modification is any operation that adds or deletes one or more
 * mappings; merely changing the value associated with an existing key is not
 * a structural modification.)  This is typically accomplished by
 * synchronizing on some object that naturally encapsulates the map.  If no
 * such object exists, the map should be "wrapped" using the
 * <tt>Collections.synchronizedMap</tt> method.  This is best done at creation
 * time, to prevent accidental unsynchronized access to the map: 
 * <pre>
 *     Map m = Collections.synchronizedMap(new TreeMap(...));
 * </pre><p>
 *
 * The iterators returned by all of this class's "collection view methods" are
 * <i>fail-fast</i>: if the map is structurally modified at any time after the
 * iterator is created, in any way except through the iterator's own
 * <tt>remove</tt> or <tt>add</tt> methods, the iterator throws a
 * <tt>ConcurrentModificationException</tt>.  Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the
 * future.
 *
 * @author  Konstantin Knizhnik
 * @version 1.01, 10/06/00
 * @see Map
 * @see HashMap
 * @see Hashtable
 * @see Comparable
 * @see Comparator
 * @see Collection
 * @see Collections#synchronizedMap(Map)
 * @since 1.2
 */

public class PersistentTreeMap extends PersistentAbstractMap
    implements SortedMap, Cloneable
{
    private Bpage root;
    private int   height;
    private Entry first;
    private Entry last;
    private PersistentComparator comparator;

    //static final int pageSize = 680; // 4Kb page
    static final int pageSize = 128;

    protected int compare(Object a, Object b) { 
	if (comparator != null) { 
	    return comparator.compare(a, b);
	} else { 
	    return ((Comparable)a).compareTo(b);
	}
    }

    static private class Entry extends Persistent implements java.util.Map.Entry { 
	Persistent key;
	Persistent value;
	Entry      prev;
	Entry      next;

        public Object getKey() { 
	    return key;
	}

        public Object getValue() { 
	    return value;
	}

	public Object setValue(Object newValue) { 
	    Object prevValue = value;
	    value = (Persistent)newValue;
	    return prevValue;
	}

        public boolean equals(Object o) { 
	    if (o instanceof Map.Entry) { 
		Map.Entry e1 = this;
		Map.Entry e2 = (Map.Entry)o;
		return (e1.getKey()==null ?
			e2.getKey()==null : e1.getKey().equals(e2.getKey()))
		    && (e1.getValue()==null ?
			e2.getValue()==null : e1.getValue().equals(e2.getValue()));
	    }
	    return false;
	}

        public int hashCode() { 
	    int h = 0;
	    if (key != null) { 
		h += key.hashCode();
	    }
	    if (value != null) { 
		h += value.hashCode();
	    }
	    return h;
	}
	
	static Metaobject defaultMetaobject = new OptimisticMetaobject();

	Entry(Persistent key, Persistent value) { 
	    this.key = key;
	    this.value = value;
	}
    }

    static private class Bpage extends Persistent { 
	Entry      leaf;
	Persistent e[]; 
	int        m;   // index of first used item at the page

	Bpage(Bpage root, Bpage brother, int pageSize) { 
	    m = pageSize-2;
	    e = new Persistent[pageSize];
	    e[pageSize-2] = brother;
	    e[pageSize-1] = root;
	    leaf = ((Bpage)root).leaf;
	}
	
	Bpage(Entry mbr, int pageSize) { 
	    m = pageSize-2;
	    e = new Persistent[pageSize];
	    e[pageSize-2] = mbr;
	    e[pageSize-1] = null;
	    leaf = null;
	}

 	Bpage(int n) {
	    e = new Persistent[n*2];
	    m = n;
	}
	
        private final int binarySearch(PersistentTreeMap tree, Object key) { 
	    int l = m, r = e.length-1;
	    while (l < r)  {
		int i = (l+r) >> 1;
		if ((e[i] instanceof Entry 
		     && tree.compare(((Entry)e[i]).key, key) < 0)
		    || (e[i] instanceof Bpage && ((Bpage)e[i]).leaf != null 
			&& tree.compare(((Bpage)e[i]).leaf.key, key) < 0))
		{
		    l = i+1; 
		} else { 
		    r = i;
		}
	    }
	    Assert.that(r == l);
	    return r;
	} 

	final Entry find(PersistentTreeMap tree, int height, Object key) {
	    int i = binarySearch(tree, key);
	    if (--height != 0) { 
		return ((Bpage)e[i]).find(tree, height, key); 
	    } else { 
		return (Entry)e[i];
	    }
	}

	final static void arraynull(Object[] a, int index, int count) { 
	    while (--count >= 0) { 
		a[index++] = null;
	    }
	}
	
	/** 
	 * Returns overflow page or old value if node with such key already exists
	 */
	final Persistent insert(PersistentTreeMap tree, int level, 
				Persistent key, Persistent value) 
	{
	    int i = binarySearch(tree, key);
	    Persistent result;
	    if (--level != 0)  {
	        result = ((Bpage)e[i]).insert(tree, level, key, value);
		if (!(result instanceof Bpage)) { 
		    return result;
		}
	    }  else { 
		Entry mbr = (Entry)e[i];
		if (mbr != null && tree.compare(mbr.key, key) == 0) { 
		    result = mbr.value;
		    mbr.value = value;
		    return result;
		} else { 
		    Entry newMbr = new Entry(key, value);
		    if (mbr == null) { 
			tree.linkLast(newMbr);
		    } else { 
			tree.linkBefore(mbr, newMbr);
		    }
		    result = newMbr;
		}		
	    } 
	    // insert before e[r]
	    if (m > 0) {
		System.arraycopy(e, m, e, m-1, i-m);
		m -= 1;
		e[i-1] = result;
		return null;
	    } else { // page is full then divide page
		int n = e.length >> 1;
		Bpage b = new Bpage(n);
		if (i < n) {
		    System.arraycopy(e, 0, b.e, n, i);
		    b.e[n+i] = result;
		    System.arraycopy(e, i, b.e, n+i+1, n-i-1);
		} else {
		    System.arraycopy(e, 0, b.e, n, n);
		    System.arraycopy(e, n, e, n-1, i-n);
		    e[i-1] = result;
		}
		b.leaf = (level != 0) 
		     ? ((Bpage)b.e[2*n-1]).leaf : (Entry)b.e[2*n-1];
		m = n-1;
		arraynull(e, 0, n-1);
		return b;
	    }
	}

	/**
	 * Returns removed value or null if key not found
	 */
	final Persistent remove(PersistentTreeMap tree, int level, Object key)
	{ 
	    int r = binarySearch(tree, key),n = e.length >> 1, last = 2*n-1;
	    Persistent result;
	    if (--level != 0)  {
		Bpage a, b;
		a = (Bpage)e[r];
	        result = a.remove(tree, level, key);
		if (a.m <= n) { 
		    return result;
		}
		// underflow of page a 
		Assert.that(a.m == n+1);
		if (r < last) { // exists greater page
		    b = (Bpage)e[r+1];
		    int bm = b.m; 
		    if (bm < n) {// reallocation of nodes between pages a and b
			int i = (n - bm + 1) >> 1;
			b.m += i;
			a.m -= i;
			System.arraycopy(a.e, n+1, a.e, n+1-i, n-1);
			System.arraycopy(b.e, bm, a.e, 2*n-i, i);
			a.leaf = (level == 1) 
			    ? (Entry)a.e[last] : ((Bpage)a.e[last]).leaf;
			arraynull(b.e, bm, i);
		    } else { // merge page a to b  
			Assert.that(bm == n); 
			b.m = 1;
			System.arraycopy(a.e, n+1, b.e, 1, n-1);
			System.arraycopy(e, m, e, m+1, r-m);
			e[m++] = null;
			// dismiss page 'a'
		    }
		} else { // page b is before a
		    b = (Bpage)e[r-1];
		    int bm = b.m; 
		    if (bm < n) { // reallocation
			int i = (n - bm + 1) >> 1;
			b.m += i;
			a.m -= i;
			System.arraycopy(b.e, 2*n-i, a.e, n+1-i, i);
			System.arraycopy(b.e, bm, b.e, bm+i, 2*n-bm-i);
			b.leaf = (level == 1) 
			    ? (Entry)b.e[last] : ((Bpage)b.e[last]).leaf;
			arraynull(b.e, bm, i);
		    } else { // merge page b to a
			Assert.that(bm == n); 
			a.m = 1;
			System.arraycopy(b.e, n, a.e, 1, n);
			System.arraycopy(e, m, e, m+1, r-1-m);
		        e[m++] = null;
			// dismiss page 'b'
		    }
		}
		if (r == last && ((Bpage)e[last]).leaf != leaf) { 
		    leaf = ((Bpage)e[last]).leaf;
		}
		return result;
	    }  else { 
		Entry mbr = (Entry)e[r];
		if (mbr != null && tree.compare(mbr.key, key) == 0) { 
		    result = mbr.value;
		    System.arraycopy(e, m, e, m+1, r-m);
		    e[m++] = null;
		    if (r == last) { 
			leaf = (Entry)e[last];
		    }
		    tree.unlink(mbr);
		    return result;
		} else { 
		    return null;
		}
	    }
	}

	static Metaobject defaultMetaobject = new OptimisticMetaobject();
    }
    
    /**
     * The number of structural modifications to the tree.
     */
    private int modCount;

    private int size;

    private final void incrementSize()   { modCount++; size++; }
    private final void decrementSize()   { modCount++; size--; }

    private final Entry successor(Entry e, int expectedModCount) 
    { 
	if (modCount != expectedModCount)
	    throw new ConcurrentModificationException();
	return e.next;
    } 

    /**
     * Test two values  for equality.  Differs from o1.equals(o2) only in
     * that it copes with with <tt>null</tt> o1 properly.
     */
    private static boolean valEquals(Object o1, Object o2) {
        return (o1==null ? o2==null : o1.equals(o2));
    }

    private final void unlink(Entry mbr) { 
	decrementSize();
	if (mbr == last) { 
	    last = last.prev;
	} else { 
	    mbr.next.prev = mbr.prev;
	}
	if (mbr == first) { 
	    first = first.next;
	} else { 
	    mbr.prev.next = mbr.next;
	}
    }

    private final void linkLast(Entry mbr) { 
	incrementSize();
	if (last == null) { 
	    first = last = mbr;
	} else { 
	    mbr.prev = last;
	    last = last.next = mbr;
	}
    }

    private final void linkBefore(Entry mbr, Entry newMbr) { 
	incrementSize();
	newMbr.prev = mbr.prev;
	newMbr.next = mbr;
	if (mbr == first) { 
	    first = newMbr;
	} else { 
	    mbr.prev.next = newMbr;
	}
	mbr.prev = newMbr;
    }
	
    /**
     * Constructs a new, empty map, sorted according to the keys' natural
     * order.  All keys inserted into the map must implement the
     * <tt>Comparable</tt> interface.  Furthermore, all such keys must be
     * <i>mutually comparable</i>: <tt>k1.compareTo(k2)</tt> must not throw a
     * ClassCastException for any elements <tt>k1</tt> and <tt>k2</tt> in the
     * map.  If the user attempts to put a key into the map that violates this
     * constraint (for example, the user attempts to put a string key into a
     * map whose keys are integers), the <tt>put(Object key, Object
     * value)</tt> call will throw a <tt>ClassCastException</tt>.
     *
     * @see Comparable
     */
    public PersistentTreeMap() {
    }

    /**
     * Constructs a new, empty map, sorted according to the given comparator.
     * All keys inserted into the map must be <i>mutually comparable</i> by
     * the given comparator: <tt>comparator.compare(k1, k2)</tt> must not
     * throw a <tt>ClassCastException</tt> for any keys <tt>k1</tt> and
     * <tt>k2</tt> in the map.  If the user attempts to put a key into the
     * map that violates this constraint, the <tt>put(Object key, Object
     * value)</tt> call will throw a <tt>ClassCastException</tt>.
     *
     * @param c the comparator that will be used to sort this map.  A
     *        <tt>null</tt> value indicates that the keys' <i>natural
     *        ordering</i> should be used.
     */
    public PersistentTreeMap(Comparator c) {
	this.comparator = (PersistentComparator)c;
    }

    /**
     * Constructs a new map containing the same mappings as the given map,
     * sorted according to the keys' <i>natural order</i>.  All keys inserted
     * into the new map must implement the <tt>Comparable</tt> interface.
     * Furthermore, all such keys must be <i>mutually comparable</i>:
     * <tt>k1.compareTo(k2)</tt> must not throw a <tt>ClassCastException</tt>
     * for any elements <tt>k1</tt> and <tt>k2</tt> in the map.  This method
     * runs in n*log(n) time.
     *
     * @param  m the map whose mappings are to be placed in this map.
     * @throws ClassCastException the keys in t are not Comparable, or
     *         are not mutually comparable.
     */
    public PersistentTreeMap(Map m) {
	putAll(m);
    }

    /**
     * Constructs a new map containing the same mappings as the given
     * <tt>SortedMap</tt>, sorted according to the same ordering.  This method
     * runs in linear time.
     *
     * @param  m the sorted map whose mappings are to be placed in this map,
     *         and whose comparator is to be used to sort this map.
     */
    public PersistentTreeMap(SortedMap m) {
        comparator = (PersistentComparator)m.comparator();
	putAll(m);
    }


    // Query Operations

    /**
     * Returns the number of key-value mappings in this map.
     *
     * @return the number of key-value mappings in this map.
     */
    public int size() {
	return size;
    }

    /**
     * Returns <tt>true</tt> if this map contains a mapping for the specified
     * key.
     *
     * @param key key whose presence in this map is to be tested.
     * 
     * @return <tt>true</tt> if this map contains a mapping for the
     *            specified key.
     * @throws ClassCastException if the key cannot be compared with the keys
     *		  currently in the map.
     * @throws NullPointerException key is <tt>null</tt> and this map uses
     *		  natural ordering, or its comparator does not tolerate
     *            <tt>null</tt> keys.
     */
    public boolean containsKey(Object key) {
	return getEntry(key) != null;
    }

    /**
     * Returns <tt>true</tt> if this map maps one or more keys to the
     * specified value.  More formally, returns <tt>true</tt> if and only if
     * this map contains at least one mapping to a value <tt>v</tt> such
     * that <tt>(value==null ? v==null : value.equals(v))</tt>.  This
     * operation will probably require time linear in the Map size for most
     * implementations of Map.
     *
     * @param value value whose presence in this Map is to be tested.
     * @since 1.2
     */
    public boolean containsValue(Object value) {
        return value==null ? valueSearchNull() : valueSearchNonNull(value);
    }

    private boolean valueSearchNull() {
	for (Entry e = first; e != null; e = e.next) { 
	    if (e.value == null) { 
		return true;
	    }
	}
	return false;
    }

    private boolean valueSearchNonNull(Object value) {
	for (Entry e = first; e != null; e = e.next) { 
	    if (value.equals(e.value)) { 
		return true;
	    }
	}
	return false;
    }

    /**
     * Returns the value to which this map maps the specified key.  Returns
     * <tt>null</tt> if the map contains no mapping for this key.  A return
     * value of <tt>null</tt> does not <i>necessarily</i> indicate that the
     * map contains no mapping for the key; it's also possible that the map
     * explicitly maps the key to <tt>null</tt>.  The <tt>containsKey</tt>
     * operation may be used to distinguish these two cases.
     *
     * @param key key whose associated value is to be returned.
     * @return the value to which this map maps the specified key, or
     *	       <tt>null</tt> if the map contains no mapping for the key.
     * @throws    ClassCastException key cannot be compared with the keys
     *		  currently in the map.
     * @throws NullPointerException key is <tt>null</tt> and this map uses
     *		  natural ordering, or its comparator does not tolerate
     *		  <tt>null</tt> keys.
     * 
     * @see #containsKey(Object)
     */
    public Object get(Object key) {
	Entry p = getEntry(key);
	return (p==null ? null : p.value);
    }

    /**
     * Returns the comparator used to order this map, or <tt>null</tt> if this
     * map uses its keys' natural order.
     *
     * @return the comparator associated with this sorted map, or
     * 	       <tt>null</tt> if it uses its keys' natural sort method.
     */
    public Comparator comparator() {
        return comparator;
    }

    /**
     * Returns the first (lowest) key currently in this sorted map.
     *
     * @return the first (lowest) key currently in this sorted map.
     * @throws    NoSuchElementException Map is empty.
     */
    public Object firstKey() {
        return first != null ? first.key : null;
    }

    /**
     * Returns the last (highest) key currently in this sorted map.
     *
     * @return the last (highest) key currently in this sorted map.
     * @throws    NoSuchElementException Map is empty.
     */
    public Object lastKey() {
        return last != null ? last.key : null;
    }

    /**
     * Returns this map's entry for the given key, or <tt>null</tt> if the map
     * does not contain an entry for the key.
     *
     * @return this map's entry for the given key, or <tt>null</tt> if the map
     * 	       does not contain an entry for the key.
     * @throws ClassCastException if the key cannot be compared with the keys
     *		  currently in the map.
     * @throws NullPointerException key is <tt>null</tt> and this map uses
     *		  natural order, or its comparator does not tolerate *
     *		  <tt>null</tt> keys.
     */
    private Entry getEntry(Object key) {
	if (root != null) { 
	    Entry e = root.find(this, height, key);
	    if (e == null) { 
		return null;
	    }
	    return e.key.equals(key) ? e : null;
	}
	return null;
    }

    /**
     * Gets the entry corresponding to the specified key; if no such entry
     * exists, returns the entry for the least key greater than the specified
     * key; if no such entry exists (i.e., the greatest key in the Tree is less
     * than the specified key), returns <tt>null</tt>.
     */
    private Entry getCeilEntry(Object key) {
	return root != null ? root.find(this, height, key) : null;
    }

    /**
     * Returns the entry for the greatest key less than the specified key; if
     * no such entry exists (i.e., the least key in the Tree is greater than
     * the specified key), returns <tt>null</tt>.
     */
    private Entry getPrecedingEntry(Object key) {
	if (root != null) { 
	    Entry e = root.find(this, height, key);
	    if (e != null && e.prev != null) { 
		return e.prev;
	    }
	}
	return null;
    }

    /**
     * Returns the key corresonding to the specified Entry.  Throw 
     * NoSuchElementException if the Entry is <tt>null</tt>.
     */
    private static Object key(Entry e) {
        if (e==null)
            throw new NoSuchElementException();
        return e.key;
    }

    /**
     * Associates the specified value with the specified key in this map.
     * If the map previously contained a mapping for this key, the old
     * value is replaced.
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * 
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.  A <tt>null</tt> return can
     *	       also indicate that the map previously associated <tt>null</tt>
     *	       with the specified key.
     * @throws    ClassCastException key cannot be compared with the keys
     *		  currently in the map.
     * @throws NullPointerException key is <tt>null</tt> and this map uses
     *		  natural order, or its comparator does not tolerate
     *		  <tt>null</tt> keys.
     */
    public Object put(Object key, Object value) {
	if (key instanceof String) { 
	    key = new ArrayOfChar((String)key);
	}
	if (root == null) { 
	    Entry e;
	    first = last = e = new Entry((Persistent)key, (Persistent)value);
	    root = new Bpage(e, pageSize);
	    height = 1;
	    incrementSize();
	    return null;
	} else { 
	    Persistent result = root.insert(this, height, 
					    (Persistent)key, (Persistent)value); 
	    if (result instanceof Bpage) { 
		root = new Bpage(root, (Bpage)result, pageSize);
		height += 1;
		return null;
	    }
	    return result;
	}
    }

    /**
     * Removes the mapping for this key from this TreeMap if present.
     *
     * @return previous value associated with specified key, or <tt>null</tt>
     *	       if there was no mapping for key.  A <tt>null</tt> return can
     *	       also indicate that the map previously associated
     *	       <tt>null</tt> with the specified key.
     * 
     * @throws    ClassCastException key cannot be compared with the keys
     *		  currently in the map.
     * @throws NullPointerException key is <tt>null</tt> and this map uses
     *		  natural order, or its comparator does not tolerate
     *		  <tt>null</tt> keys.
     */
    public Object remove(Object key) {
	if (root == null) { 
	    return null;
	} else { 
	    Persistent result = root.remove(this, height, key);
	    if (root.m == pageSize-1) { 
		if (--height == 0) { 
		    root = null;
		} else { 
		    root = (Bpage)root.e[pageSize-1];
		}
	    }
	    return result;
	}
    }

    private final void deleteEntry(Entry e, int expectedModCount) { 
	if (modCount != expectedModCount)
	    throw new ConcurrentModificationException();
	remove(e.key);
    }

    private final void deleteEntry(Entry e) { 
	remove(e.key);
    }

    /**
     * Removes all mappings from this TreeMap.
     */
    public void clear() {
	modCount++;	
	size = 0;
	root = null;
	height = 0;
	first = last = null;
    }

    /**
     * Returns a shallow copy of this <tt>TreeMap</tt> instance. (The keys and
     * values themselves are not cloned.)
     *
     * @return a shallow copy of this Map.
     */
    public Object clone() {
        PersistentTreeMap clone = null;
	try { 
	    clone = (PersistentTreeMap)super.clone();
	} catch (CloneNotSupportedException e) { 
	    throw new InternalError();
	}

        // Put clone into "virgin" state (except for comparator)
        clone.root = null;
        clone.size = 0;
        clone.height = 0;
        clone.first = null;
        clone.last = null;
        clone.modCount = 0;
        clone.keySet = clone.entrySet = null;
        clone.values = null;

        // Initialize clone with our mappings
	clone.putAll(this);
        return clone;
    }


    // Views

    /**
     * These fields are initialized to contain an instance of the appropriate
     * view the first time this view is requested.  The views are stateless,
     * so there's no reason to create more than one of each.
     */
    private transient Set		keySet = null;
    private transient Set		entrySet = null;
    private transient Collection	values = null;

    /**
     * Returns a Set view of the keys contained in this map.  The set's
     * iterator will return the keys in ascending order.  The map is backed by
     * this <tt>PersistentTreeMap</tt> instance, so changes to this map are reflected in
     * the Set, and vice-versa.  The Set supports element removal, which
     * removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Set.remove</tt>, <tt>removeAll</tt>,
     * <tt>retainAll</tt>, and <tt>clear</tt> operations.  It does not support
     * the <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a set view of the keys contained in this PersistentTreeMap.
     */
    public Set keySet() {
	if (keySet == null) {
	    keySet = new AbstractSet() {
		public java.util.Iterator iterator() {
		    return new Iterator(KEYS);
		}

		public int size() {
		    return PersistentTreeMap.this.size();
		}

                public boolean contains(Object o) {
                    return containsKey(o);
                }

		public boolean remove(Object o) {
                    int oldSize = size;
                    PersistentTreeMap.this.remove(o);
		    return size != oldSize;
		}

		public void clear() {
		    PersistentTreeMap.this.clear();
		}
	    };
	}
	return keySet;
    }

    /**
     * Returns a collection view of the values contained in this map.  The
     * collection's iterator will return the values in the order that their
     * corresponding keys appear in the tree.  The collection is backed by
     * this <tt>PersistentTreeMap</tt> instance, so changes to this map are reflected in
     * the collection, and vice-versa.  The collection supports element
     * removal, which removes the corresponding mapping from the map through
     * the <tt>Iterator.remove</tt>, <tt>Collection.remove</tt>,
     * <tt>removeAll</tt>, <tt>retainAll</tt>, and <tt>clear</tt> operations.
     * It does not support the <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a collection view of the values contained in this map.
     */
    public Collection values() {
	if (values == null) {
            values = new PersistentAbstractCollection() {
                public java.util.Iterator iterator() {
                    return new Iterator(VALUES);
                }

                public int size() {
                    return PersistentTreeMap.this.size();
                }

                public boolean contains(Object o) {
                    for (Entry e = first; e != null; e = e.next) { 
                        if (valEquals(e.value, o))
                            return true;
		    }
                    return false;
                }

		public boolean remove(Object o) {
                    for (Entry e = first; e != null; e = e.next) {
                        if (valEquals(e.value, o)) {
                            deleteEntry(e);
                            return true;
                        }
                    }
                    return false;
		}

                public void clear() {
                    PersistentTreeMap.this.clear();
                }
            };
        }
        return values;
    }

    /**
     * Returns a set view of the mappings contained in this map.  The set's
     * iterator returns the mappings in ascending key order.  Each element in
     * the returned set is a <tt>Map.Entry</tt>.  The set is backed by this
     * map, so changes to this map are reflected in the set, and vice-versa.
     * The set supports element removal, which removes the corresponding
     * mapping from the PersistentTreeMap, through the <tt>Iterator.remove</tt>,
     * <tt>Set.remove</tt>, <tt>removeAll</tt>, <tt>retainAll</tt> and
     * <tt>clear</tt> operations.  It does not support the <tt>add</tt> or
     * <tt>addAll</tt> operations.
     *
     * @return a set view of the mappings contained in this map.
     * @see Map.Entry
     */
    public Set entrySet() {
	if (entrySet == null) {
	    entrySet = new AbstractSet() {
                public java.util.Iterator iterator() {
                    return new Iterator(ENTRIES);
                }

                public boolean contains(Object o) {
                    if (!(o instanceof Map.Entry))
                        return false;
                    Map.Entry entry = (Map.Entry)o;
                    Object value = entry.getValue();
                    Entry p = getEntry(entry.getKey());
                    return p != null && valEquals(p.value, value);
                }

		public boolean remove(Object o) {
                    if (!(o instanceof Map.Entry))
                        return false;
                    Map.Entry entry = (Map.Entry)o;
                    Object value = entry.getValue();
                    Entry p = getEntry(entry.getKey());
                    if (p != null && valEquals(p.value, value)) {
                        deleteEntry(p);
                        return true;
                    }
                    return false;
                }

                public int size() {
                    return PersistentTreeMap.this.size();
                }

                public void clear() {
                    PersistentTreeMap.this.clear();
                }
            };
        }
	return entrySet;
    }

    /**
     * Returns a view of the portion of this map whose keys range from
     * <tt>fromKey</tt>, inclusive, to <tt>toKey</tt>, exclusive.  (If
     * <tt>fromKey</tt> and <tt>toKey</tt> are equal, the returned sorted map
     * is empty.)  The returned sorted map is backed by this map, so changes
     * in the returned sorted map are reflected in this map, and vice-versa.
     * The returned sorted map supports all optional map operations.<p>
     *
     * The sorted map returned by this method will throw an
     * <tt>IllegalArgumentException</tt> if the user attempts to insert a key
     * less than <tt>fromKey</tt> or greater than or equal to
     * <tt>toKey</tt>.<p>
     *
     * Note: this method always returns a <i>half-open range</i> (which
     * includes its low endpoint but not its high endpoint).  If you need a
     * <i>closed range</i> (which includes both endpoints), and the key type
     * allows for calculation of the successor a given key, merely request the
     * subrange from <tt>lowEndpoint</tt> to <tt>successor(highEndpoint)</tt>.
     * For example, suppose that <tt>m</tt> is a sorted map whose keys are
     * strings.  The following idiom obtains a view containing all of the
     * key-value mappings in <tt>m</tt> whose keys are between <tt>low</tt>
     * and <tt>high</tt>, inclusive:
     * 	    <pre>    SortedMap sub = m.submap(low, high+"\0");</pre>
     * A similar technique can be used to generate an <i>open range</i> (which
     * contains neither endpoint).  The following idiom obtains a view
     * containing all of the key-value mappings in <tt>m</tt> whose keys are
     * between <tt>low</tt> and <tt>high</tt>, exclusive:
     * 	    <pre>    SortedMap sub = m.subMap(low+"\0", high);</pre>
     *
     * @param fromKey low endpoint (inclusive) of the subMap.
     * @param toKey high endpoint (exclusive) of the subMap.
     * 
     * @return a view of the portion of this map whose keys range from
     * 	       <tt>fromKey</tt>, inclusive, to <tt>toKey</tt>, exclusive.
     * 
     * @throws ClassCastException if <tt>fromKey</tt> and <tt>toKey</tt>
     *         cannot be compared to one another using this map's comparator
     *         (or, if the map has no comparator, using natural ordering).
     * @throws IllegalArgumentException if <tt>fromKey</tt> is greater than
     *         <tt>toKey</tt>.
     * @throws NullPointerException if <tt>fromKey</tt> or <tt>toKey</tt> is
     *	       <tt>null</tt> and this map uses natural order, or its
     *	       comparator does not tolerate <tt>null</tt> keys.
     */
    public SortedMap subMap(Object fromKey, Object toKey) {
	return new SubMap(fromKey, toKey);
    }

    /**
     * Returns a view of the portion of this map whose keys are strictly less
     * than <tt>toKey</tt>.  The returned sorted map is backed by this map, so
     * changes in the returned sorted map are reflected in this map, and
     * vice-versa.  The returned sorted map supports all optional map
     * operations.<p>
     *
     * The sorted map returned by this method will throw an
     * <tt>IllegalArgumentException</tt> if the user attempts to insert a key
     * greater than or equal to <tt>toKey</tt>.<p>
     *
     * Note: this method always returns a view that does not contain its
     * (high) endpoint.  If you need a view that does contain this endpoint,
     * and the key type allows for calculation of the successor a given key,
     * merely request a headMap bounded by <tt>successor(highEndpoint)</tt>.
     * For example, suppose that suppose that <tt>m</tt> is a sorted map whose
     * keys are strings.  The following idiom obtains a view containing all of
     * the key-value mappings in <tt>m</tt> whose keys are less than or equal
     * to <tt>high</tt>:
     * <pre>
     *     SortedMap head = m.headMap(high+"\0");
     * </pre>
     *
     * @param toKey high endpoint (exclusive) of the headMap.
     * @return a view of the portion of this map whose keys are strictly
     * 	       less than <tt>toKey</tt>.
     *
     * @throws ClassCastException if <tt>toKey</tt> is not compatible
     *         with this map's comparator (or, if the map has no comparator,
     *         if <tt>toKey</tt> does not implement <tt>Comparable</tt>).
     * @throws IllegalArgumentException if this map is itself a subMap,
     *         headMap, or tailMap, and <tt>toKey</tt> is not within the
     *         specified range of the subMap, headMap, or tailMap.
     * @throws NullPointerException if <tt>toKey</tt> is <tt>null</tt> and
     *	       this map uses natural order, or its comparator does not
     *	       tolerate <tt>null</tt> keys.
     */
    public SortedMap headMap(Object toKey) {
	return new SubMap(toKey, true);
    }

    /**
     * Returns a view of the portion of this map whose keys are greater than
     * or equal to <tt>fromKey</tt>.  The returned sorted map is backed by
     * this map, so changes in the returned sorted map are reflected in this
     * map, and vice-versa.  The returned sorted map supports all optional map
     * operations.<p>
     *
     * The sorted map returned by this method will throw an
     * <tt>IllegalArgumentException</tt> if the user attempts to insert a key
     * less than <tt>fromKey</tt>.<p>
     *
     * Note: this method always returns a view that contains its (low)
     * endpoint.  If you need a view that does not contain this endpoint, and
     * the element type allows for calculation of the successor a given value,
     * merely request a tailMap bounded by <tt>successor(lowEndpoint)</tt>.
     * For For example, suppose that suppose that <tt>m</tt> is a sorted map
     * whose keys are strings.  The following idiom obtains a view containing
     * all of the key-value mappings in <tt>m</tt> whose keys are strictly
     * greater than <tt>low</tt>: <pre>
     *     SortedMap tail = m.tailMap(low+"\0");
     * </pre>
     *
     * @param fromKey low endpoint (inclusive) of the tailMap.
     * @return a view of the portion of this map whose keys are greater
     * 	       than or equal to <tt>fromKey</tt>.
     * @throws ClassCastException if <tt>fromKey</tt> is not compatible
     *         with this map's comparator (or, if the map has no comparator,
     *         if <tt>fromKey</tt> does not implement <tt>Comparable</tt>).
     * @throws IllegalArgumentException if this map is itself a subMap,
     *         headMap, or tailMap, and <tt>fromKey</tt> is not within the
     *         specified range of the subMap, headMap, or tailMap.
     * @throws NullPointerException if <tt>fromKey</tt> is <tt>null</tt> and
     *	       this map uses natural order, or its comparator does not
     *	       tolerate <tt>null</tt> keys.
     */
    public SortedMap tailMap(Object fromKey) {
	return new SubMap(fromKey, false);
    }

    private class SubMap extends PersistentAbstractMap implements SortedMap {
        /**
         * fromKey is significant only if fromStart is false.  Similarly,
         * toKey is significant only if toStart is false.
         */
        private boolean fromStart = false, toEnd = false;
	private Object  fromKey,           toKey;

	SubMap(Object fromKey, Object toKey) {
	    if (compare(fromKey, toKey) > 0)
		throw new IllegalArgumentException("fromKey > toKey");
	    this.fromKey = fromKey;
	    this.toKey = toKey;
	}

	SubMap(Object key, boolean headMap) {
            compare(key, key); // Type-check key

            if (headMap) {
                fromStart = true;
                toKey = key;
            } else {
                toEnd = true;
                fromKey = key;
            }
	}

	SubMap(boolean fromStart, Object fromKey, boolean toEnd, Object toKey){
            this.fromStart = fromStart;
            this.fromKey= fromKey;
            this.toEnd = toEnd;
            this.toKey = toKey;
	}

	public boolean isEmpty() {
	    return entrySet.isEmpty();
	}

	public boolean containsKey(Object key) {
	    return inRange(key) && PersistentTreeMap.this.containsKey(key);
	}

	public Object get(Object key) {
	    if (!inRange(key))
                return null;
	    return PersistentTreeMap.this.get(key);
	}

	public Object put(Object key, Object value) {
	    if (!inRange(key))
		throw new IllegalArgumentException("key out of range");
	    return PersistentTreeMap.this.put(key, value);
	}

        public Comparator comparator() {
            return comparator;
        }

        public Object firstKey() {
            Object obj = key(fromStart ? first:getCeilEntry(fromKey));
            if (!toEnd && compare(obj, toKey) >= 0)
                throw(new NoSuchElementException());
            return obj;
        }

        public Object lastKey() {
            Object obj = key(toEnd ? last : getPrecedingEntry(toKey));
            if (!fromStart && compare(obj, fromKey) < 0)
                throw(new NoSuchElementException());
            return obj;
        }

	private transient Set entrySet = new EntrySetView();

	public Set entrySet() {
	    return entrySet;
	}

	private class EntrySetView extends AbstractSet {
	    private transient int size = -1, sizeModCount;

	    public int size() {
		if (size == -1 || sizeModCount != PersistentTreeMap.this.modCount) {
		    size = 0;  sizeModCount = PersistentTreeMap.this.modCount;
		    java.util.Iterator i = iterator();
		    while (i.hasNext()) {
			size++;
			i.next();
		    }
		}
		return size;
	    }

	    public boolean isEmpty() {
		return !iterator().hasNext();
	    }

	    public boolean contains(Object o) {
		if (!(o instanceof Map.Entry))
		    return false;
		Map.Entry entry = (Map.Entry)o;
		Object key = entry.getKey();
                if (!inRange(key))
                    return false;
                PersistentTreeMap.Entry node = getEntry(key);
                return node != null &&
                       valEquals(node.value, entry.getValue());
	    }

	    public boolean remove(Object o) {
		if (!(o instanceof Map.Entry))
		    return false;
		Map.Entry entry = (Map.Entry)o;
		Object key = entry.getKey();
                if (!inRange(key))
                    return false;
                PersistentTreeMap.Entry node = getEntry(key);
                if (node!=null && valEquals(node.value,entry.getValue())){
		    deleteEntry(node);
		    return true;
                }
                return false;
	    }

	    public java.util.Iterator iterator() {
		return new Iterator(
                    (fromStart ? first : getCeilEntry(fromKey)),
                    (toEnd     ? null  : getCeilEntry(toKey)));
	    }
	}

        public SortedMap subMap(Object fromKey, Object toKey) {
            if (!inRange2(fromKey))
		throw new IllegalArgumentException("fromKey out of range");
            if (!inRange2(toKey))
                throw new IllegalArgumentException("toKey out of range");
            return new SubMap(fromKey, toKey);
        }

        public SortedMap headMap(Object toKey) {
            if (!inRange2(toKey))
                throw new IllegalArgumentException("toKey out of range");
            return new SubMap(fromStart, fromKey, false, toKey);
        }

        public SortedMap tailMap(Object fromKey) {
            if (!inRange2(fromKey))
                throw new IllegalArgumentException("fromKey out of range");
            return new SubMap(false, fromKey, toEnd, toKey);
        }

	private boolean inRange(Object key) {
	    return (fromStart || compare(key, fromKey) >= 0) &&
                   (toEnd     || compare(key, toKey)   <  0);
	}

        // This form allows the high endpoint (as well as all legit keys)
	private boolean inRange2(Object key) {
	    return (fromStart || compare(key, fromKey) >= 0) &&
                   (toEnd     || compare(key, toKey)   <= 0);
        }
    }

    // Types of Iterators
    private static final int KEYS = 0;
    private static final int VALUES = 1;
    private static final int ENTRIES = 2;

    /**
     * PersistentTreeMap Iterator.
     */
    private class Iterator implements java.util.Iterator {
	private int type;
	private int expectedModCount = PersistentTreeMap.this.modCount;
	private Entry lastReturned = null;
	private Entry next;
	private Entry firstExcluded = null;

	Iterator(int type) {
	    this.type = type;
	    next = first;
	}

	Iterator(Entry first, Entry firstExcluded) {
	    type = ENTRIES;
	    next = first;
	    this.firstExcluded = firstExcluded;
	}

	public boolean hasNext() {
	    return next != firstExcluded;
	}

	public Object next() {
	    if (next == firstExcluded)
		throw new NoSuchElementException();
	    lastReturned = next;
	    next = successor(next, expectedModCount);
	    return (type == KEYS ? lastReturned.key :
		    (type == VALUES ? lastReturned.value : lastReturned));
	}

	public void remove() {
	    if (lastReturned == null)
		throw new IllegalStateException();
	    deleteEntry(lastReturned, expectedModCount);
	    expectedModCount++;
	    lastReturned = null;
	}
    }
}





